/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_fdata.h"
#include "delay.h"
#include "minrtsFftUtil.h"
#define clight  2.9979250e8 /*m per sec */
#define twopi   6.2831853072
#define NFFTGRISM 1024
#define NFFTPRISM 512
/*#define WAVENOMARGIN 4.e5*/
#define WAVENOMARGIN 8.e5
#define SMOOTHSIGMA  0.2 /* 100 ms default */
#define MAXOPDGRISM 200 /* maximum findable OPD with GRISM */
#define MAXOPDPRISM 200 /* maximum findable OPD with PRISM */
#define MGRAD 2  /* max number of delay gradients to try */
#define PGRAD 1.0 /* size of gradiant (rad/sigma) */
#define ENDDROP 0.1  /* fraction pts at ends of delayogram to drop */

/*#define SIGMAOUT 0.125*/ /* 8 readouts per sigma */
#define SIGMAOUT 0.25 /* 4 readouts per sigma */
#define EXCLUDED  1.2 /* exclude this # * SIGMAOUT frames on each side of center of
                         smoothing interval to suppress biases */
/*#define DEBUG 1*/


void printUsage()
{
  printf ("\n Usage: oirFGroupDelay DataFile DelayFile \n");
  printf ("[-smooth smooth] [-medsmooth] medSmooth [-search] [-maxopd maxOpd] [-first firstGuessFile] [-ngrad nGrad] [-exinterval exInt] [-sigout sigout]\n\n");
  printf("  DataFile:    filename of output of oirRotateInsOpd \n");
  printf("  DelayFile:   filename of output from this program \n");
  printf("  smooth:      gaussian 1-sigma smoothing interval (in sec)");
  printf("     when combining spectra; defaults to 100 msec \n");
  printf("  medsmooth:   median smooth output delay fn by this inverval (seconds)\n");
  printf("  -search: search mode\n");
  printf("  maxOpd: restrict opd search to maxOpd of either tracking modal OPD or track specified in firstGuess\n");
  printf("  firstGuessFile: file containing first guess delay table.\n");
  printf("  nGrad        2*nGrad+1 = number of opd grandients to try. Default=1\n");
  printf("  sigout: compress OPD output in time by sampling only at \n");
  printf("     intervals of this fraction of smoothing interval (=smooth)\n");
  printf("     default: 0.25 of smooth\n");
  printf("  exInt: to avoid induced noise bias, the estimate of OPD at time t\n");
  printf("     excludes data taken within exInt*sigout*smooth of t\n");
  printf("     default: 1.2*sigout*smooth\n");
  exit(1);
}
/* ************************************************************************
 * Find the index and value of the maximum and mininum of a float array
 * Inputs:
 *         a    array of input values 
 *         n    number of input values
 * Outputs:
 *         *min value of mininum of array
 *         *max value of maximum of array
 *         *imin position (0-rel) of minimum value in array
 *         *imax position (0-rel) of maximum value in array
 *   
 *********************************************************************** */
void mmax(float *a, int n, float *min, float *max, int *imin, int *imax) {
   int i,j,k;
   float in,ax;
   j=0;
   k=0;
   in = a[0];
   ax = in;
   for (i=0;i<n;i++){
      if (a[i] > ax) {
         ax = a[i];
         j  = i;
      } 
      if (a[i] < in) {
         in = a[i];
         k  = i;
      } 
   }
   *imin = k;
   *imax = j;
   *min  = in;
   *max  = ax;
   return;
}

/*
 *     add a new value (in) to a sorted list (list[nList])
*/    
void addSortedD(double *list, int nList, double in) {
   int i,j;
   int nL;
   i=0;
   nL = nList-1;
   while (i < nL &&(in > list[i])) i++;
   for (j=nL;j>i;j--) list[j] = list[j-1];
   list[i] = in;
   return;
}
/*
 *     remove a  value (out) to a sorted list (list[nList])
*/    
void removeSortedD(double *list, int nList, double out) {
   int i,j;
   int nL;
   i=0;
   nL = nList-1;
   while (i < nL&&(out != list[i])) i++;
   for(j=i;j<nL;j++) list[j] = list[j+1];
   return;
}
/*
 *     combine the previous two operations
*/    
void addRemoveSortedD(double *list, int nList, double in, double out) {
   int iout, iin, i;
   if (out == in) return;
   iout=0;
   while (iout < nList &&(out != list[iout])) iout++;
   if (in < out) { 
      iin = 0;
      while (iin < iout &&(in > list[iin])) iin++;
      for (i=iout;i>iin;i--) list[i] = list[i-1];
   } else {
      iin = iout;
      while (iin < (nList-1) &&(in > list[iin+1])) iin++;
      for (i=iout;i<iin;i++) list[i] = list[i+1];
   }
   list[iin] = in;
   return;
}
void minrtsMedianSmoothD(double *data, int nData, int nSmooth, double* outData) {
   double *mData;
   int nOdd, n2;
   int i,nList, l2;
   int out, odd;

   odd = nSmooth - 2 * (nSmooth/2); /* I can only take the median of an odd number of pts */
   if (odd) nOdd = nSmooth; else nOdd = nSmooth+1;
   n2 = (int)(nOdd/2);
   mData = (double *)malloc(nOdd * sizeof(double)); /* work array to store current samples */

   for (i=0;i<=n2;i++) addSortedD(mData,i+1,data[i]); /* initial load of mData */
   nList = n2;
   for (out=0; out<n2;out++) {
      l2           = (int)(nList/2);
      nList++;
      outData[out] = mData[l2];
      addSortedD(mData, nList+1, data[nList]);
   }
   for (out=n2; out<(nData-n2-1); out++) {
      outData[out] = mData[n2];
      addRemoveSortedD(mData, nOdd, data[out+n2+1], data[out-n2]);
   }
   outData[nData-n2-1] = mData[n2];
   nList = n2;
   for (out=nData-n2; out<nData; out++) {
      removeSortedD(mData, n2+nList+1, data[nData-nList-2]);
      l2           = (int)(nList/2);
      outData[out] = mData[l2];
      nList--;
   }
   free(mData); 
   return;
}
/* **************************************************************************
 *  Sum the rows of a buffer with relative weights exp(-0.5*((iRow-oRow)/sigma))^2)
 *  INPUTS:
 *         buffer[2*nY+1][nX]   array of input rows
 *         inRows[2*nY+1]       array of input row positions, one for each row
 *         oRow                 "central" input row position
 *         nX                   pixels per row
 *         nY                   number of rows is 2*nY + 1
 *         sigma                standard deviation of weighting function in rows
 *         smooth[nX]           output row
 ************************************************************************** */
float smoothDelayData (float *buffer, int *inRows,  double* inTime,
   double outTime, float exTime, double *weightedTime, 
   int nX, int nY, float sigma, float pGrad, float *waveNo, float *smooth) {

   int i,j,k;
   int nComplex;
   float weight, totalWeight;
   float x,c,s,w;
   float wave10;
   float dTime;
   int   noExtime;
   double wTime;

   wave10 = twopi*1.e5; /* waveno for 10 microns */
   noExtime = (exTime <= 0.);
   nComplex = (int)(nX/2);

   totalWeight = 0.;  /* total of weights */
   wTime       = 0.;  /* total of weights */
   for (j=0;j<nX;j++) smooth[j] = 0.;/* zero accumulator */

   for (i=0; i<nY; i++) if (inRows[i] > 0){
      dTime = inTime[i] - outTime;
/*      if (((dTime < exTime) && (dTime > (-exTime))) || noExtime) { */
      if (((dTime > exTime) || (dTime < (-exTime))) || noExtime) { 
         x = dTime/sigma; 
         weight = exp(-0.5* x*x);
         w = pGrad*x/wave10;
         totalWeight = totalWeight + weight;
         wTime       = wTime + weight*dTime; /* weighted time difference */
         for (j=0;j<nComplex;j++) {
            k = 2*j+i*nX;
            c = cos(waveNo[j]*w);
            s = sin(waveNo[j]*w);
            smooth[2*j] = smooth[2*j] + weight * 
               (c*(*(buffer+k)) - s*(*(buffer+k+1)));
            smooth[2*j+1] = smooth[2*j+1] + weight * 
               (s*(*(buffer+k)) + c*(*(buffer+k+1)));
         }
      }
   }
   if (totalWeight > 0.) {
      for (j=0;j<nX;j++) smooth[j] = smooth[j]/totalWeight;
      *weightedTime = outTime + wTime/totalWeight;
   }
   return totalWeight;
}
/* **************************************************************************
 *    compute absolute value of delay function, 
 *    find peak, do parabolic interpolation to subpixel accuracy
 *    around peak, and return position and value of peak
 *  INPUTS:
 *      fft        fft object containing coordinate and pixel information
 *      delayData  complex delay function data
 *  OUTPUTS:
 *      mx         value of interpolated peak of absolute value of delayData
 *  RETURNS:
 *                 position of interpolated peak of absolute value, in units specified by fft
 *      
 ************************************************************************** */
double getGroupOpd(minrtsFft1d *fft, float *delayData, float *mx, int nBegin, int nEnd){
   float nX;
   float mn;
   int ix, in;
   int i;
   float ix2;

   nX = fft->nX;
/* 
      compute square of abs value and move to 1st half of array 
*/
   for (i=0;i<nX;i++) delayData[i] = sqrt(delayData[2*i]*delayData[2*i]
      +delayData[2*i+1]*delayData[2*i+1]);
   mmax (delayData+nBegin, nEnd-nBegin, &mn, mx, &in, &ix); /* min max and position */
   ix = ix + nBegin;
   if (ix < 3 || (ix > (nX-4))) ix2 = (double)fft->kGrid[ix]; 
      else ix2 = minrtsQPk(7, fft->kGrid+ix-3, delayData+ix-3, mx, 0);
   if (ix2 < fft->kGrid[ix-3]) ix2 = fft->kGrid[ix-3];
   if (ix2 > fft->kGrid[ix+3]) ix2 = fft->kGrid[ix+3];
    return (double)ix2;

} /* end of get Group Opd */

/*
 *            for a given time, look through the table of groupDelay versus
 *            time, find the two nearest time points, and interpolate
 *            between them to estimate the groupDelay.  Keep track
 *            of the last point found, in order to speed up searching
*/ 
double interpolateGroupDelay(double time, long nDelay, double *groupDelay, 
   double *delayTime, int *lastFound) {

   long  iDelay;
   float x;

   iDelay = *lastFound;
   if (iDelay < 0) iDelay = 0;

   while ((iDelay < (nDelay-1)) && (delayTime[iDelay+1] < time)) ++iDelay;
   x = (time-delayTime[iDelay])/(delayTime[iDelay+1]-delayTime[iDelay]);
   *lastFound = iDelay;
   return (1-x)*groupDelay[iDelay] + x*groupDelay[iDelay+1];
} /*  end of interpolateGroupDelay */
int main (int argc, char** argv) {

   int   i, j, ierr;
   float x;
   int   nfft, nfft2, nffthalf;
   long  nRow, iRow, oRow, pRow ;
   long  nTableRows;
   float totalWeight;
   int   nx, nx2 ;
   float *frequency, *waveNo;
   float minWave, maxWave;

   char  outFileName[130];
   char *bang="!";
   char* infile_name     = NULL;
   char* outfile_name    = NULL;
   char* firstfile_name = NULL;
   fitsfile *inFile;
   fitsfile *outFile = NULL;

   char  gris[12]; /*disperser name */
   float detdit; /*exposure time*/
   double fullTime; /*total frame time*/
   char  history[80];

   int    iKey, maxstep;
   double opd,  opd0;
   float  opd1, opd2, opdScan, oldOpdScan;
   float  opdMin, opdMax;
   float  *rotOpd;
   float  maxOpd;
   int nEnd, nBegin, nDrop, nSearch;

   imaging_fdata_table_def    *idata;
   imaging_fdata_table_def    *odata;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_row          *inData;
   imaging_fdata_row          *outData;

   delay_table_row *delayData;
   delay_table_def *dTable;
   double          *delayTimes;
   double          *groupDelays;
   float           *delayAmp;
   double          *firstGroupDelays;
   double          *firstDelayTimes;
   float           *firstDelayAmp;
   int             lastTableDelay;
   long            nFirstDelays;
   float           exTime;
   float           sigmaOut;

   float delayPk[1+2*MGRAD];
   float ampPk[1+2*MGRAD];
   double weightedTime[1+2*MGRAD];
   float mn;
   int   imax, imin;
   int   NGRAD;

   minrtsFft1d *groupFft;
   float *work1, *work2, *work3, *work4;
   float gridParms[3];

   float  *smBuffer, *smBufferOut;
   double *smBufferTime, startTime, delTime, outTime, maxTime;
   double minBufferTime;
   int    smBufferWidth; 
   float  smBW2;
   int    *smBufferI;
   int    smI;
   float  opt_smooth;

   int     doMedian;
   float   medTime ;
   int     medFrames;
   double *medianBuffer;

   int doSearch; /* search mode rather than track mode */

   int iarg;  /* for parsing input */

   ierr       = 0;
   doSearch   = 0;
   NGRAD      = 1;
   doMedian   = 0;
   medTime    = -1.e9;


   /**
    * Parse the command line options 
    */
   if (argc < 3) {
      printUsage();
      return 1;
   }
   printf("      ****** oirFGroupDelay ******\n");
   infile_name  = argv[1];
   outfile_name = argv[2];
   printf ("   Using input from    %s\n", infile_name);
   printf ("   Putting output in   %s\n", outfile_name);
   maxOpd       = -1.e9; /* large value */

   opt_smooth = SMOOTHSIGMA; /* default */
   exTime     = EXCLUDED; /*default*/
   sigmaOut   = SIGMAOUT; /*default*/

   iarg = 3; 
   while (iarg < argc ) {
      if (!strcmp("-smooth", argv[iarg])) sscanf(argv[++iarg],"%f",&opt_smooth); 
      if (!strcmp("-search", argv[iarg])) doSearch = 1;
      if (!strcmp("-maxopd", argv[iarg])) sscanf(argv[++iarg],"%f",&maxOpd); 
      if (!strcmp("-ngrad",  argv[iarg])) sscanf(argv[++iarg],"%i",&NGRAD); 
      if (!strcmp("-medsmooth", argv[iarg])) sscanf(argv[++iarg],"%f",&medTime); 
      if (!strcmp("-first",  argv[iarg])) firstfile_name=argv[++iarg];
      if (!strcmp("-exinterval",  argv[iarg])) sscanf(argv[++iarg],"%f",&exTime); 
      if (!strcmp("-sigout",  argv[iarg])) sscanf(argv[++iarg],"%f",&sigmaOut); 
      ++iarg;
   }
   /* printf ("   Smoothing over                 %-8.0f millisec\n",1.e3*opt_smooth);
   printf ("   outputTime/sigma=              %-8.3f millisec\n",sigmaOut); */
   if (doSearch == 1) printf("   Operating in SEARCH mode.\n");
      else printf("   Operating in TRACK mode.\n");
   delTime  = sigmaOut*opt_smooth/86400; /* output every eighth smooth in days */
   printf ("   outputInterval               = %8.3f sec\n",delTime*86400);
   exTime   = delTime*exTime;     /* default */
   printf ("   Excluding %.1f * outputInterval == %-.3f sec.\n",exTime/delTime, exTime*86400.);
/*
      is there a first file?  If so read it in.
*/
   if (firstfile_name != NULL) {
      printf("   Using firstfile %s\n",firstfile_name);
      fits_open_file (&inFile, firstfile_name, READONLY, &ierr);
      if (ierr != 0) {
         printf("ERROR:fits open first file of %s gave error %i\n",firstfile_name,
            ierr);
      printf("%s \n",strerror(errno));
         return 1;
       }

      dTable = open_delay_table(inFile, &ierr);
      if (ierr != 0) {
         printf("error %i opening delay table \n", ierr);
         return 1;
      }
      delayData = make_delay_table_row(dTable);
      i= fits_get_num_rows(inFile, &nFirstDelays, &ierr);
      firstGroupDelays = (double *)malloc(nFirstDelays*sizeof(double));;
      firstDelayTimes  = (double *)malloc(nFirstDelays*sizeof(double));;
      firstDelayAmp    = (float *)malloc(nFirstDelays*sizeof(float));;
      for (iRow =1; iRow<=nFirstDelays;iRow++) {
         i=read_delay_table(inFile, dTable, delayData, iRow, &ierr);
         if (ierr != 0) {
            printf("error %i reading delay table row %li\n", ierr, iRow);
            return 1;
         }
         firstGroupDelays[iRow-1] = delayData->delay*clight; /*meters*/
         firstDelayAmp[iRow-1]    = delayData->amplitude; 
         firstDelayTimes[iRow-1]  = delayData->time;   /* mjd */
      }
/*
      close file and clean up structures
*/
      i = close_delay_table(inFile, dTable, &ierr);
   }    /*  done processing first guess table */
/*
      open input file and get detector info
*/
   fits_open_file (&inFile, infile_name, READONLY, &ierr);
   if (ierr != 0) {
      printf("ERROR:fits openfile of %s gave error %i\n",infile_name,
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS GRIS NAME", gris,0,&ierr);
   printf("   disperser                    = %s\n", gris);
   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO DET DIT", &detdit,0,&ierr);
   printf("   exposure time (DIT)          = %8.3f sec\n", detdit);
   printf("   coherent smoothing sigma     = %8.3f sec ==%8.1f frames\n", opt_smooth, opt_smooth/detdit);
   if (medTime < 0) printf("   NOT median smoothing of the group delay.\n");
      else {
         doMedian = 1;
         printf("   median smoothing width       = %8.3f sec ==%8.1f outputIntervals\n",
            medTime, round(medTime/delTime/86400.));
      }
   printf("   number of linear opd fits    = %i \n",2*NGRAD+1);
   if (NGRAD > MGRAD) {
      printf (" NGRAD is larger than MGRAD; quitting!\n");
      return 1;
   }
/*
 *             set default values of maxOpd if not specified 
*/
   if (maxOpd < (-1.e8)) 
      if (gris[0] == 'G') maxOpd = MAXOPDGRISM; else maxOpd = MAXOPDPRISM;
   
   printf("   maximum search opd           = %8.1f micron\n", maxOpd);
   maxOpd = 1.e-6*maxOpd; /* meters */

/*
      create output data table 
*/
   strcpy(outFileName, bang);
   strcat(outFileName, outfile_name);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      printf ("ERROR: opening fits output %s %i\n",outfile_name, ierr);
      printf("%s \n",strerror(errno));
      return 1;
      }
   i = fits_copy_header(inFile, outFile, &ierr);
   snprintf(history,80, "oirFGroupDelay");
   i = fits_write_history(outFile, history, &ierr);
   snprintf(history,80, "INFILE %s", infile_name);
   i = fits_write_history(outFile, history, &ierr);
   i = fits_write_key(outFile, TSTRING, "DELMODE", (void*)"TIME",
      "delay mode is time not frames", &ierr);
   i = fits_write_key(outFile, TFLOAT, "OSMOOTH", (void*)&opt_smooth, 
      "Smoothing halfwidth in oirFGroupDelay", &ierr);

   DetectorDef = read_imaging_detector_table(inFile, &ierr);
   nx  = DetectorDef->naxis[0][0];
   nx2 = nx/2;
/*       
      get frequency information from detector table
      screw around with naxis; this is necessary
      because data in .insopd file is "pseudocomplex"
      so nx is twice what it is supposed to be.
*/
   DetectorDef->naxis[0][0] = nx2;
   frequency = imaging_detector_wcs(DetectorDef, 1);
   waveNo = (float *)malloc(nx2 * sizeof(float));
   for (i=0; i<nx2; i++) waveNo[i] = twopi * frequency[i]/clight;
   rotOpd = (float *)malloc(nx * sizeof(float)); /* phase rotation in search mode */
/*
      configure data table and get MAXSTEP
*/
   idata  = open_imaging_fdata_table(inFile, &ierr);
   iKey   = table_find_key(idata->def, "MAXSTEP");
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&maxstep,
      &ierr);
/*    get number of rows and full frame time */
   i       = fits_get_num_rows (inFile,  &nRow, &ierr); 
   printf("   total number of frames       = %ld\n", nRow);
   inData  = make_imaging_fdata_row(idata);
   i = read_imaging_fdata_table(inFile, idata, inData, 1, &ierr);
   fullTime = inData->time;
   i = read_imaging_fdata_table(inFile, idata, inData, 2, &ierr);
   fullTime = inData->time-fullTime; /*frame time including dead time in days*/
   nTableRows = (long)(3.0*nRow * fullTime/delTime); /* include fudge factor */
   if (nTableRows < nRow) nTableRows = nRow;
   printf("   full cycle time              = %8.3f sec\n", fullTime*8.64e4);
/*
/* 
      pick nfft and
      configure smoothing buffer
*/
   if (gris[0] == 'G') nfft = NFFTGRISM; else nfft = NFFTPRISM;

   nfft2         = 2* nfft;
   nffthalf      = nfft/2;
   smBW2         = opt_smooth/86400.; /* days */
   smBufferWidth = (int)(4. * opt_smooth/fullTime/86400); /* points */
   smBuffer      = (float *)malloc(smBufferWidth * nx *sizeof(float));
   smBufferOut   = (float *)malloc(nx * sizeof(float));
   smBufferI     = (int *)malloc(smBufferWidth *sizeof(int));
   smBufferTime  = (double *)malloc(smBufferWidth *sizeof(double));
   memset(smBufferI, 0, smBufferWidth*sizeof(int));
   memset(smBufferTime, 0, smBufferWidth*sizeof(double));
   
/*     
   row size(viewed as complex); alter it for gridded output
*/
   DetectorDef->naxis[0][0] = nfft2;
   odata = make_imaging_fdata_table_def (outFile, 
      DetectorDef, maxstep, &ierr); 
   if (ierr !=0) {
      printf ("ERROR: creating data definition\n");
      return 1;
      }
   outData = make_imaging_fdata_row(odata);
/*    setup FFT objects and work arrays */
   mmax(waveNo, nx2, &minWave, &maxWave, &i, &j); 
   gridParms[0]  = minWave - WAVENOMARGIN;
   gridParms[1]  = (maxWave - minWave + 2.*WAVENOMARGIN)/nfft;
   gridParms[2]  = nfft;
   groupFft      = minrtsMakeFft1d(gridParms, 0);

   opd1          = -groupFft->kGrid[0]; /* maximum relative OPD in output array; doesn't include opd0 */
   opd2          = groupFft->kGrid[0] - groupFft->kGrid[1];/* negativedelta opd between output pts of fft */
/*                  select search range if no first guess specified */
   
   nDrop    = (int)(ENDDROP*nfft); /* pixels to skip at beginning of Delay-O-Gram in any case */
   nSearch  = (int)(maxOpd/opd2);  /* 1/2 of #pixels to search as specified by maxopd */
   if (nSearch < 0) nSearch = -nSearch;
   nBegin   = nffthalf - nSearch;
   if (nBegin < nDrop) nBegin = nDrop;
   nEnd = (nfft/2)-(int)(maxOpd/opd2);
   j    = (int)(ENDDROP*nfft);
   if (nEnd < j) nEnd = j;

   work1 = (float *)malloc(nx * sizeof(float));
   work2 = (float *)malloc(nfft2 * sizeof(float));
   work3 = (float *)malloc(nx * sizeof(float));
   work4 = (float *)malloc(nfft2 * sizeof(float));
/*       
         get OPD offset value from header
*/
   iKey = table_find_key(idata->def, "OPD0");
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&opd0, &ierr);
   opdScan = opd0; /* initialize to harmless values */
   opdMin  = opd0 + 0.5 * groupFft->kGrid[0];
   opdMax  = opd0 - 0.5 * groupFft->kGrid[0];

/*
         set up tables to calculate and output group delays
*/
   delayTimes  = (double *)malloc(nTableRows * sizeof(double));
   groupDelays = (double *)malloc(nTableRows * sizeof(double));
   delayAmp    = (float *) malloc(nTableRows * sizeof(float));
/*
         now loop over input data, 
         output DATA is just FFT of input data (wrt wavenumber) with
         most table data just copied.
         DELAY data is peak of absolute value of FFTs of coherently
         averaged frames with gaussian weighting wrt time difference
         from nominal output times.  These are not calculated every frame
         but only at intervals delTime
*/ 
   smI            = 0; /* index in rolling smooth buffer to store next input data*/
   oRow           = 0; /* index in output arrays to store delay data */
   lastTableDelay = 0; /* no delays yet found in table */
   outTime  = delTime; /* set time corresponding to first output delay point */
   for (iRow=1; iRow<=nRow; iRow++) {
      i = read_imaging_fdata_table(inFile, idata, inData, iRow, &ierr);
/*
            skip invalid data
*/
      if (!isnan(inData->time)) {
/*
            copy nonData info
*/
      if (iRow == 1) startTime = inData->time;
/*            get lots of output points */
      outData->time   = inData->time;
      outData->opd[0] = inData->opd[0];
      outData->opd[1] = inData->opd[1];
      outData->local_opd[1] = inData->local_opd[1];
      outData->local_opd[0] = inData->local_opd[0];
      outData->exp = inData->exp;
      outData->offset[0] = inData->offset[0];
      outData->offset[1] = inData->offset[1];
      outData->rotation  = inData->rotation;
      outData->frame     = inData->frame;
      outData->step      = inData->step;
      outData->reference = inData->reference;
      outData->targtype[0] = inData->targtype[0];  
/*
        if search mode, put back opd0 - opdScan in phase
        change center of FFT opd range to current opd if it falls
        outside the current search range.

        If you do this, zap the current smoother buffer by setting row
        values to large negative numbers
*/
      if (doSearch) {
         opd = inData->opd[1]+inData->local_opd[1] - inData->opd[0] - inData->local_opd[0];
         if ((opd > opdMax) || (opd < opdMin) || (iRow == 1)) {
/*
        rotate already stored points to new phase center
*/
            oldOpdScan = opdScan;
            opdScan = opd0 - opd;
            for (j=0; j<nx2; j++) {
               rotOpd[2*j]   = cos((opdScan-oldOpdScan) * waveNo[j]);
               rotOpd[2*j+1] = sin((opdScan-oldOpdScan) * waveNo[j]);
            }
            for (i=0; i< smBufferWidth; i++) {
               for (j=0; j<nx2; j++) {
                  x = inData->data[0][2*j]*rotOpd[2*j] - 
                     inData->data[0][2*j+1]*rotOpd[2*j+1];
                  inData->data[0][2*j+1] = inData->data[0][2*j]*rotOpd[2*j+1] +
                     inData->data[0][2*j+1]*rotOpd[2*j];
                  inData->data[0][2*j] = x;
               }
            } /* end reposition old data */
/*
         now calculate rotations from opd0 to new phase center
*/
            for (j=0; j<nx2; j++) {
               rotOpd[2*j]   = cos(opdScan * waveNo[j]);
               rotOpd[2*j+1] = sin(opdScan * waveNo[j]);
            }
            opdMin  = opd + 0.5 * groupFft->kGrid[0];
            opdMax  = opd - 0.5 * groupFft->kGrid[0];
         }  /* end of if out of fft range */
/*
         rotate new point to current search center
*/
         for (j=0; j<nx2; j++) {
            x = inData->data[0][2*j]*rotOpd[2*j] - 
               inData->data[0][2*j+1]*rotOpd[2*j+1];
            inData->data[0][2*j+1] = inData->data[0][2*j]*rotOpd[2*j+1] +
               inData->data[0][2*j+1]*rotOpd[2*j];
            inData->data[0][2*j]   = x;
         }  /* end of rotate new data to search center */
      }  /* end if doSearch */
/*
        do complex ffts and output into DATA table
*/
      
      minrtsFftGridFftC(groupFft, nx2, waveNo, inData->data[0], 0, outData->data[0]); 
 
      i = write_imaging_fdata_table(outFile, odata, outData, iRow, &ierr);
      if (ierr != 0) {
         printf("ERROR:output error at line %li %i\n", iRow, ierr);
         return 1;
      }

/*
         copy into smoothing buffer in order to calculate group delays
*/
      memcpy (smBuffer+smI*nx, inData->data[0], nx*sizeof(float));
      smBufferI[smI]    = iRow;
      maxTime           = outData->time - startTime; /*days*/
      minBufferTime     = smBufferTime[smI]; /* time of last point removed from buffer or 0*/
      smBufferTime[smI] = maxTime;
      if (++smI == smBufferWidth) smI = 0;  /* prepare buffer index for next input point */
/*
         is it time to calculate group delay?
         do it if the time of the last input frame is more than 2sigma
         later than the nominal output time
*/
      if (maxTime >= outTime + 2.*smBW2) { 
/*
         if first guess file has been specified, recalculate nBegin and nEnd
         to follow first guess +/- maxOpd
*/
         if (firstfile_name != NULL) {
            opd = interpolateGroupDelay(outTime+startTime, nFirstDelays, 
               firstGroupDelays, firstDelayTimes, &lastTableDelay);
            opd = opd - opd0;  /*delay relative to opd0 (meters)  */
            nBegin = nffthalf + (int)(opd/opd2) - nSearch;
            if (nBegin < nDrop) nBegin = nDrop;
            if (nBegin > (nfft-nDrop-2*nSearch)) nBegin = nfft-nDrop-2*nSearch;
            nEnd = nBegin + 2*nSearch;
            if (nEnd > (nfft-nDrop)) nEnd = nfft-nDrop;
         }

/*
         sum with gaussian weights into averaging buffer
         and assumed linear rate of change of OPD during interval
         Take FFT from wavenumber->delay space, then find which
         of the rate of change values yields the largest fringe peak
*/
         for (i=0;i<(1+2*NGRAD);i++) {
            totalWeight = smoothDelayData (smBuffer, smBufferI,  smBufferTime,
               outTime, exTime, weightedTime+i, 
               nx, smBufferWidth, smBW2, PGRAD*(i-NGRAD), waveNo, smBufferOut);
            minrtsFftGridFftC(groupFft, nx2, waveNo, smBufferOut, 0, outData->data[0]);
            delayPk[i] = getGroupOpd(groupFft, outData->data[0], ampPk+i, nBegin, nEnd);
         }
         if (totalWeight > 0.1) {
            mmax(ampPk, 1+2*NGRAD, &mn, delayAmp+oRow, &imin, &imax);
            groupDelays[oRow] = delayPk[imax];
            delayTimes[oRow]  = weightedTime[imax] + startTime;
            if (doSearch) groupDelays[oRow] = groupDelays[oRow]+opdScan;
            oRow++;
         }
         outTime = outTime + delTime; 
         while(outTime < minBufferTime) outTime = outTime + delTime; 
      }  /* end of new output necessary */
      j = iRow/100;
      if ((iRow-j*100)==0) fprintf(stderr,"Frame %ld of %ld\r",iRow, nRow);
   }  /* test if data valid */
   }  /*  loop over input rows */
/*
         process last output rows
*/
   while (outTime <= maxTime) {
         for (i=0;i<(1+2*NGRAD);i++) {
            totalWeight = smoothDelayData (smBuffer, smBufferI,  smBufferTime,
               outTime, exTime, weightedTime+i,
               nx, smBufferWidth, smBW2, PGRAD*(i-NGRAD), waveNo, smBufferOut);
            minrtsFftGridFftC(groupFft, nx2, waveNo, smBufferOut, 0, outData->data[0]);
            delayPk[i] = getGroupOpd(groupFft, outData->data[0], ampPk+i, nBegin, nEnd);
         }
         if (totalWeight > 0.1) {
            mmax(ampPk, 1+2*NGRAD, &mn, delayAmp+oRow, &imin, &imax);
            groupDelays[oRow] = delayPk[imax];
            delayTimes[oRow]  = weightedTime[imax]+startTime;
            if (doSearch) groupDelays[oRow] = groupDelays[oRow]+opdScan;
            oRow++;
         }
         outTime = outTime + delTime; 
         while(outTime < minBufferTime) outTime = outTime + delTime; 
   }
   printf("   number of output delays      = %li\n", oRow);
   i = fits_write_key(outFile, TDOUBLE, "OPD0", (void*)(&opd0), "OFFSET TO OPD", &ierr);
   i = fits_write_key(outFile, TFLOAT,  "OPD1", (void*)(&opd1), " ", &ierr);
   i = fits_write_key(outFile, TFLOAT,  "OPD2", (void*)(&opd2), " ", &ierr);
   i = fits_write_key(outFile, TFLOAT,  "GRIDPAR1", (void*)(gridParms), "WaveNo Grid Zero Point", &ierr);
   i = fits_write_key(outFile, TFLOAT,  "GRIDPAR2", (void*)(1+gridParms), "WaveNo Grid Delta", &ierr);
   kill_imaging_fdata_table_def(idata);
   kill_imaging_fdata_table_def(odata);
   kill_imaging_fdata_row(inData);
   kill_imaging_fdata_row(outData);

/*
 *    if requested, median smooth delay data before output
 *    and copy back into groupDelay variable
*/
   if (doMedian) {
      if (delTime > 0) medFrames = round(medTime/delTime/86400.); else
         medFrames = 3;
      medianBuffer = (double *)malloc(oRow*sizeof(double));
      minrtsMedianSmoothD(groupDelays, oRow-1, medFrames, medianBuffer);
      memcpy(groupDelays, medianBuffer, (oRow-1)*sizeof(double));
      free(medianBuffer);
   }
/*
       calculate and output estimates of group delay
*/

   dTable = make_delay_table_def(outFile, "1776-07-04", &ierr);
   delayData = make_delay_table_row(dTable);
   for (iRow=1;iRow<oRow;iRow++) {
      delayData->time      = delayTimes[iRow-1];
      delayData->telescope = 2;
      delayData->amplitude = delayAmp[iRow-1];
      delayData->delay     = (opd0-groupDelays[iRow-1])/clight; /* in seconds */
      i = write_delay_table(outFile, dTable, delayData, iRow, &ierr);
   }
   kill_delay_table_def(dTable);
   kill_delay_table_row(delayData);
   fits_close_file(inFile, &ierr);
   fits_close_file(outFile, &ierr);
   free(frequency);
   free(waveNo);
   free(work1);
   free(work2);
   free(work3);
   free(work4);
   free(delayTimes);
   free(delayAmp);
   free(groupDelays);
   free(smBuffer);
   free(smBufferI);
   free(smBufferOut);
   free(smBufferTime);
   free(rotOpd);
   if (firstfile_name != NULL) {
      free(firstGroupDelays);
      free(firstDelayTimes);
      free(firstDelayAmp);
   }

   minrtsKillFft1d(groupFft);
   return 0;
}
